'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

EXISTENCE = "existence"
EXACTLY_ONE = "exactly_one"
INIT = "init"
RESPONDED_EXISTENCE = "responded_existence"
RESPONSE = "response"
PRECEDENCE = "precedence"
SUCCESSION = "succession"
ALTRESPONSE = "altresponse"
ALTPRECEDENCE = "altprecedence"
ALTSUCCESSION = "altsuccession"
CHAINRESPONSE = "chainresponse"
CHAINPRECEDENCE = "chainprecedence"
CHAINSUCCESSION = "chainsuccession"
ABSENCE = "absence"
COEXISTENCE = "coexistence"
NONCOEXISTENCE = "noncoexistence"
NONSUCCESSION = "nonsuccession"
NONCHAINSUCCESSION = "nonchainsuccession"
